--[[

Slither Link - a Slither Link/Puzzle Loop/Loop the Loop game.
Made entirely by celynwalters, 2013 (except for the font :3)

I designed this from scratch, myself, because I love this puzzle. I was frustrated because
if you wanted to play this, you either had to print one out (wasteful), play on a mobile device
(mine is broken, app is buggy) or use a computer (the pointing and clicking is tedious and non-optimal).

Plus it was a learning exercise (steep learning curve!), and I can customise it to however I like.
And löve2D is so good, you can run it on nearly anything.

Things to know (i.e. Not the game rules. I wouldn't write them HERE.):
	- The game grid is contained in map{}
		- 4 means empty space
		- All other numbers are shown (0,1,2,3)
	- The lines you put in are contained in solution{}
		- The key is sequentially the tops of cells, then bottoms, then sides (left to right)
		- 0 means nothing
		- 1 means line
		- 2 means cross
	- The loop status is in loop{}
		- 3 means being checked in the current loop (should not be observed)
		- 2 means not a loop
		- 1 means a loop
		- 0 means haven't checked yet (should not be onserved)
	- ftr_something generally means "first time run"

To sort out:
	- Lines that reach dead ends are not thought about...

To add:
	- Generate puzzles! (Not a clue how to do yet...)
		- Include difficulty choice
	- Menus!
		- Customise keys!
		- Sound options!
			- Make music for the game!
		- Display options!
			- Enable zoom in...? (/Auto resize)
			- Colour schemes!
	- Times!
		- Highscores!
	- Updateable!
	- Smooth moving cursor
	- A proper background
																		]]
--======================================================================--
-- Shortcuts/Libraries													--
--======================================================================--
gfx = love.graphics
lfs = love.filesystem
li = love.image
keydown = love.keyboard.isDown
audio = love.audio
mouse = love.mouse
mousedown = love.mouse.isDown
function sin(x) return math.sin(math.rad(x)) end
function cos(x) return math.cos(math.rad(x)) end
function tan(x) return math.tan(math.rad(x)) end
function asin(x) return math.deg(math.asin(x)) end
function acos(x) return math.deg(math.acos(x)) end
function atan(x) return math.deg(math.atan(x)) end

require "extra"
--======================================================================--
-- The 'Big Three' - load, update, draw									--
--======================================================================--
function love.load()
-------- System stuff
	resolutions = gfx.getModes()
	table.sort(resolutions, function(a,b) return a.width*a.height < b.width*b.height end)   -- sort from smallest to largest
	screen_w,screen_h,fullscreen,vsync,fsaa = gfx.getMode()
	gfx.setBackgroundColor(60,60,60)
-------- Resources
--	sound_music = audio.newSource("music.mp3","stream")
	sound_line = audio.newSource("line.ogg","static")
	sound_cross = audio.newSource("cross.ogg","static")
	sound_cancel = audio.newSource("cancel.ogg","static")
	fontsize = 50
	font = gfx.newFont("tf2build.ttf",fontsize)
	debug_font = gfx.newFont(14)
--	sound_music:setLooping("true")
-------- Setup values
	debug = true			-- By default
	initiate()
	gridsize = fontsize
	gfx.setPointSize(5)
	selected = 0
	siblings = 0
	completed = false
	offset_x = (screen_w/2)-(map_w/2*gridsize)
	offset_y = 50
	remember = {}
	reset_loop(0)
	loops = 0
	checkloop(1,1)
	undo = {}
	toggle_lines = false
	toggle_crosses = false
	toggle_loops = false
	toggle_music = false
--------
end
--------------------------------------------------------------------------
function love.update(dt)
-------- FPS limiter
	if dt < 1/60 then
		love.timer.sleep(1/60 - dt)
	end
--------
	mouse_x, mouse_y = mouse.getPosition()

	selected_x = math.floor((mouse_x-offset_x)/gridsize)+1
	selected_y = math.floor((mouse_y-offset_y)/gridsize)+1

	-- Keeps user inputted lines to a space (0), line(1) or cross(2)
	for i=1,#solution do
		if solution[i] > 2 then solution[i] = 0 end
		if solution[i] < 0 then solution[i] = 2 end
	end

	if toggle_music == true then
		if ftr_music == nil then
			audio.play(sound_music)
			ftr_music = false
		end
	else
--		audio.pause(sound_music)
		ftr_music = nil
	end
--------
end
--------------------------------------------------------------------------
function love.draw()
--------
	-- Draw numbers
	gfx.setFont(font)
	gfx.setColor(color_white)
	for i=1,map_h do
		for j=1,map_w do
			if map[i][j] ~= 4 then
				gfx.printf(map[i][j],(gridsize*(j-1))+gridsize/2+offset_x,gridsize*(i-1)+offset_y,10,"center")
			end
		end
	end
------ Loop status
	if toggle_loops == true then
		for i=1,loop_h do
			for j=1,loop_w do
				if loop[i][j] == 1 then
					gfx.setColor(0,0,255,100)
					gfx.rectangle("fill",gridsize*j-gridsize+offset_x,gridsize*(i-1)+offset_y,gridsize,gridsize)
				elseif loop[i][j] == 2 then
					gfx.setColor(100,100,255,100)
					gfx.rectangle("fill",gridsize*j-gridsize+offset_x,gridsize*(i-1)+offset_y,gridsize,gridsize)					
				end
--				gfx.printf(loop[i][j],(gridsize*(j-1))+gridsize/2+offset_x,gridsize*(i-1)+offset_y,10,"center")
-- KEY!!
			end
		end
	end
	-- Draw dots
	gfx.setColor(color_white)
	for i=0,map_h do
		for j=0,map_w do
			gfx.point(gridsize*j+offset_x,gridsize*i+offset_y)
		end
	end
	-- Draw selection where mouse is
	gfx.setColor(255,255,255,100)
	if selected_x > 0 and selected_x <= map_w then
		if selected_y > 0 and selected_y <= map_h then
			gfx.rectangle("fill",gridsize*selected_x-gridsize+offset_x,gridsize*(selected_y-1)+offset_y,gridsize,gridsize)
		end
	end
	-- Draw user lines (top and bottoms)
	gfx.setColor(color_white)
	for i=1,map_w*(map_h+1) do
		if solution[i] == 1 then
			draw_lines_tops(i)
		end
	end
	-- Draw user lines (sides)
	gfx.setColor(color_white)
	for i=map_w*(map_h+1)+1,#solution do
		local t = i-(map_w*(map_h+1)+1)
		if solution[i] == 1 then
			draw_lines_sides(t,i)
		end
	end
	-- Draw crosses (top and bottoms)
	gfx.setColor(color_red)
	for i=1,map_w*(map_h+1) do
		if solution[i] == 2 then
			draw_crosses_tops(i)
		end
	end
	-- Draw crosses (sides)
	gfx.setColor(color_red)
	for i=map_w*(map_h+1)+1,#solution do
		local t = i-(map_w*(map_h+1)+1)
		if solution[i] == 2 then
			draw_crosses_sides(t,i)
		end
	end
	-- Draws coloured rectangles
	for i=1,map_h do
		for j=1,map_w do
			if toggle_lines == true then
				if count1(i,j,map[i][j]) == "more" then
					gfx.setColor(255,0,0,100)
					gfx.rectangle("fill",gridsize*j-gridsize+offset_x,gridsize*(i-1)+offset_y,gridsize,gridsize)
				elseif count1(i,j,map[i][j]) == "match" then
					gfx.setColor(0,255,0,100)
					gfx.rectangle("fill",gridsize*j-gridsize+offset_x,gridsize*(i-1)+offset_y,gridsize,gridsize)
				end
			end
			if toggle_crosses == true then
				if count2(i,j,map[i][j]) == "crosses" then
					gfx.setColor(255,255,0,100)
					gfx.rectangle("fill",gridsize*j-gridsize+offset_x,gridsize*(i-1)+offset_y,gridsize,gridsize)
				end
			end
		end
	end


-------- Debug mode items to display (SHOULD BE DRAWN LAST)
	gfx.setFont(debug_font)
	gfx.setColor(color_white)
	gfx.print("Debug: "..tostring(debug),0,14*1-14)
	if debug == true then
		gfx.print("Map dimensions: "..map_w.."x"..map_h,0,14*2-14)
		gfx.print("Mouse position: "..mouse_x..", "..mouse_y,0,14*3-14)
		gfx.print("offset_x: "..offset_x,0,14*4-14)
		gfx.print("offset_y: "..offset_y,0,14*5-14)
		gfx.print("selected_x: "..selected_x,0,14*6-14)
		gfx.print("selected_y: "..selected_y,0,14*7-14)
		gfx.print("Completed: "..tostring(completed),0,14*8-14)
		gfx.print("#loops: "..loops,0,14*9-14)
		gfx.print("#undos: "..#undo,0,14*10-14)
		gfx.print("Display line status: "..tostring(toggle_lines),0,14*11-14)
		gfx.print("Display cross status: "..tostring(toggle_crosses),0,14*12-14)
		gfx.print("Display loop status: "..tostring(toggle_loops),0,14*13-14)
	end
--------
end
--======================================================================--
-- Everything else														--
--======================================================================--
function love.keypressed(key)
--------
	if key == "`" then
		debug = not debug
	end

	if key == "m" then
--		toggle_music = not toggle_music
	end

	if key == "l" then
		toggle_lines = not toggle_lines
	end

	if key == "c" then
		toggle_crosses = not toggle_crosses
	end

	if key == "o" then
		toggle_loops = not toggle_loops
	end

	if (selected_x > 0) and (selected_x <= map_w) and (selected_y > 0) and (selected_y <= map_h) then
		if key == "w" then
			local a = (((selected_y-1)*map_w)+selected_x)
			solution[a] = solution[a] + 1
			table.insert(undo,a)
			if solution[a] == 1 then audio.stop(sound_line) audio.play(sound_line)
			elseif solution[a] == 2 then audio.stop(sound_cross) audio.play(sound_cross)
			elseif solution[a] == 3 then audio.stop(sound_cancel) audio.play(sound_cancel)
			end
			checkloop(1,1)
		end
		if key == "s" then
			local a = (((selected_y)*map_w)+selected_x)
			table.insert(undo,a)
			solution[a] = solution[a] + 1
			if solution[a] == 1 then audio.stop(sound_line) audio.play(sound_line)
			elseif solution[a] == 2 then audio.stop(sound_cross) audio.play(sound_cross)
			elseif solution[a] == 3 then audio.stop(sound_cancel) audio.play(sound_cancel)
			end
			checkloop(1,1)
		end
		if key == "a" then
			local a = (map_w)*(map_h+1)+(map_w+1)*(selected_y-1)+selected_x
			table.insert(undo,a)
			solution[a] = solution[a] + 1
			if solution[a] == 1 then audio.stop(sound_line) audio.play(sound_line)
			elseif solution[a] == 2 then audio.stop(sound_cross) audio.play(sound_cross)
			elseif solution[a] == 3 then audio.stop(sound_cancel) audio.play(sound_cancel)
			end
			checkloop(1,1)
		end
		if key == "d" then
			local a = (map_w)*(map_h+1)+(map_w+1)*(selected_y-1)+selected_x+1
			table.insert(undo,a)
			solution[a] = solution[a] + 1
			if solution[a] == 1 then audio.stop(sound_line) audio.play(sound_line)
			elseif solution[a] == 2 then audio.stop(sound_cross) audio.play(sound_cross)
			elseif solution[a] == 3 then audio.stop(sound_cancel) audio.play(sound_cancel)
			end
			checkloop(1,1)
		end
		if key == "u" then
			if #undo > 0 then
				solution[undo[#undo]] = solution[undo[#undo]] - 1
				table.remove(undo,#undo)
			end
			checkloop(1,1)
		end
	end
--------
end
--------------------------------------------------------------------------
function count1(incols,inrows,number)	-- Returns status of number of lines around cells
	local sides = 0
	-- Tops
	if solution[((map_w)*(incols-1))+(inrows)] == 1 then sides = sides + 1 end
	-- Bottoms
	if solution[((map_w)*(incols))+(inrows)] == 1 then sides = sides + 1 end
	-- Lefts
	if solution[(map_h)*(map_w)+(inrows-1)+((incols)*(map_w+1))] == 1 then sides = sides + 1 end
	-- Rights
	if solution[(map_h)*(map_w)+(inrows)+((incols)*(map_w+1))] == 1 then sides = sides + 1 end
	-- Return answer
	if sides > number then return "more" end
	if sides < number then return "fewer" end
	if (sides == number) and (number ~= 4) then return "match" end
	if number == 4 then return "fewer" end
end
--------------------------------------------------------------------------
function count2(incols,inrows,number)	-- Returns status of number of crosses around cells
	local sides = 0
	-- Tops
	if solution[((map_w)*(incols-1))+(inrows)] == 2 then sides = sides + 1 end
	-- Bottoms
	if solution[((map_w)*(incols))+(inrows)] == 2 then sides = sides + 1 end
	-- Lefts
	if solution[(map_h)*(map_w)+(inrows-1)+((incols)*(map_w+1))] == 2 then sides = sides + 1 end
	-- Rights
	if solution[(map_h)*(map_w)+(inrows)+((incols)*(map_w+1))] == 2 then sides = sides + 1 end
	-- Return answer
	if (number ~= 4) and (4-sides < number) then return "crosses" end
end
--------------------------------------------------------------------------
function checkloop(x,y)
	if ftr_checkloop == nil then
		reset_loop(0)
		ftr_checkloop = 1
		is_a_loop = true
		loops = 0
	end
	up = true			-- Means that the loop can continue in that direction.
	down = true			-- 	These will turn false if they need to.
	left = true
	right = true
	if solution[(map_w*(y-1))+(x)] == 1 then up = false end						-- If there is a line above the cell
	if solution[(map_w*y)+(x)] == 1 then down = false end						-- If there is a line below the cell
	if solution[(map_h*map_w)+(x-1)+(y*(map_w+1))] == 1 then left = false end	-- If there is a line left of the cell
	if solution[(map_h*map_w)+x+(y*(map_w+1))] == 1 then right = false end	-- If there is a line right of the cell
	if (y == 1) and (up == true) then			-- If the loop can go too high
		is_a_loop = false
		up = false
	end
	if (y == loop_h) and (down == true) then	-- If the loop can go too low
		is_a_loop = false
		down = false
	end
	if (x == 1) and (left == true) then			-- If the loop can go too far left
		is_a_loop = false
		left = false
	end
	if (x == loop_w) and (right == true) then	-- If the loop can go too far right
		is_a_loop = false
		right = false
	end
	if (up == true) and (loop[y-1][x] == 0) then			-- Process can only check one path at a time,
		if (down == true) and (loop[y+1][x] == 0) then		--	so it must remember junctions.
			remember_down(x,y)
		end
		if (left == true) and (loop[y][x-1] == 0) then
			remember_left(x,y)
		end
		if (right == true) and (loop[y][x+1] == 0) then
			remember_right(x,y)
		end
		loop[y][x] = 3
		y = y-1
		checkloop(x,y)
		do return end
	elseif (down == true) and (loop[y+1][x] == 0) then
		if (left == true) and (loop[y][x-1] == 0) then
			remember_left(x,y)
		end
		if (right == true) and (loop[y][x+1] == 0) then
			remember_right(x,y)
		end
		loop[y][x] = 3
		y = y+1
		checkloop(x,y)
		do return end
	elseif (left == true) and (loop[y][x-1] == 0) then
		if (right == true) and (loop[y][x+1] == 0) then
			remember_right(x,y)
		end
		loop[y][x] = 3
		x = x-1
		checkloop(x,y)
		do return end
	elseif (right == true) and (loop[y][x+1] == 0) then
		loop[y][x] = 3
		x = x+1
		checkloop(x,y)
		do return end
	else
		if #remember > 0 then
			loop[y][x] = 3
			y = remember[#remember-1]
			x = remember[#remember]
			table.remove(remember,#remember)
			table.remove(remember,#remember)
			checkloop(x,y)
			do return end
		else
			loop[y][x] = 3
			fill_squares()
			find_new_loop(x,y)
		end
	end
	ftr_checkloop = nil
	-- Convenient to put this here
	if loops == 1 then
		completed = true
	else
		completed = false
	end
	for i=1,map_h do
		for j=1,map_w do
			if map[j][i] ~= 4 then
				if count1(j,i,map[j][i]) ~= "match" then
					completed = false
				end
			end
		end
	end

end
--------------------------------------------------------------------------
function remember_down(x,y)
	loop[y+1][x] = 3
	table.insert(remember,y+1)
	table.insert(remember,x)
end
function remember_left(x,y)
	loop[y][x-1] = 3
	table.insert(remember,y)
	table.insert(remember,x-1)
end
function remember_right(x,y)
	loop[y][x+1] = 3
	table.insert(remember,y)
	table.insert(remember,x+1)
end
--------------------------------------------------------------------------
function fill_squares()
	for i=1,loop_h do
		for j=1,loop_w do
			if loop[j][i] == 3 then
				if is_a_loop == true then
					loop[j][i] = 1
				else
					loop[j][i] = 2
				end
			end
		end
	end
end
--------------------------------------------------------------------------
function find_new_loop(x,y)
	if is_a_loop == true then loops = loops + 1 end
	for i=1,loop_h do
		for j=1,loop_w do
			if loop[j][i] == 0 then
				is_a_loop = true
				checkloop(i,j)
			end
		end
	end
end
--------------------------------------------------------------------------
function draw_lines_tops(i)
	if math.floor(i/map_w) == i/map_w then
		gfx.line(
			(i+(map_w)-(map_w*(math.floor(i/map_w))))*gridsize-gridsize+offset_x,
			(-1-map_w+math.floor(map_w+i/map_w))*gridsize+offset_y,
			gridsize+(i+(map_w)-(map_w*(math.floor(i/map_w))))*gridsize-gridsize+offset_x,
			(-1-map_w+math.floor(map_w+i/map_w))*gridsize+offset_y)
	else
		gfx.line(
			(i-(map_w*(math.floor(i/map_w))))*gridsize-gridsize+offset_x,
			math.floor(i/map_w)*gridsize+offset_y,
			gridsize+(i-(map_w*(math.floor(i/map_w))))*gridsize-gridsize+offset_x,
			math.floor(i/map_w)*gridsize+offset_y)
	end
end
--------------------------------------------------------------------------
function draw_lines_sides(t,i)
	gfx.line(
		gridsize*(t-(math.floor(t/(map_w+1))*(map_w+1)))+offset_x,
		gridsize*math.floor(t/(map_w+1))+offset_y,
		gridsize*(t-(math.floor(t/(map_w+1))*(map_w+1)))+offset_x,
		gridsize*math.floor(t/(map_w+1))+gridsize+offset_y)
end
--------------------------------------------------------------------------
function draw_crosses_tops(i)
	if math.floor(i/map_w) == i/map_w then
		gfx.point(
			gridsize*(i-(map_w*(math.floor(i/(map_w+1)))))-gridsize+gridsize/2+offset_x,
			gridsize*(math.floor(map_w+i/map_w)-map_w-1)+offset_y)
	else
		gfx.point(
			gridsize*(i-(map_w*(math.floor(i/map_w))))-gridsize+gridsize/2+offset_x,
			gridsize*(-map_w+math.floor(map_w+i/map_w))+offset_y)
	end
end
--------------------------------------------------------------------------
function draw_crosses_sides(t,i)
	gfx.point(
		gridsize*(t-(math.floor(t/(map_w+1))*(map_w+1)))+offset_x,
		gridsize*math.floor(t/(map_w+1))+gridsize/2+offset_y)
end
--------------------------------------------------------------------------
--------------------------------------------------------------------------
